# Copyright (C) 2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import inspect

import pytest

from swh.search import get_search
from swh.search.api.client import RemoteSearch
from swh.search.elasticsearch import ElasticSearch
from swh.search.in_memory import InMemorySearch
from swh.search.interface import SearchInterface

SEARCH_IMPLEMENTATIONS_KWARGS = [
    ("remote", RemoteSearch, {"url": "localhost"}),
    (
        "elasticsearch",
        ElasticSearch,
        {"hosts": ["localhost"], "indexes": {"origin": {"index": "test"}}},
    ),
]

SEARCH_IMPLEMENTATIONS = SEARCH_IMPLEMENTATIONS_KWARGS + [
    ("memory", InMemorySearch, None),
]


def test_get_search_failure():
    with pytest.raises(ValueError, match="Unknown search class"):
        get_search("unknown-search")


@pytest.mark.parametrize("class_,expected_class,kwargs", SEARCH_IMPLEMENTATIONS)
def test_get_search(mocker, class_, expected_class, kwargs):
    mocker.patch("swh.search.elasticsearch.Elasticsearch")
    if kwargs:
        concrete_search = get_search(class_, **kwargs)
    else:
        concrete_search = get_search(class_)
    assert isinstance(concrete_search, expected_class)


@pytest.mark.parametrize("class_,expected_class,kwargs", SEARCH_IMPLEMENTATIONS_KWARGS)
def test_get_search_deprecation_warning(mocker, class_, expected_class, kwargs):
    with pytest.warns(DeprecationWarning):
        concrete_search = get_search(class_, args=kwargs)
    assert isinstance(concrete_search, expected_class)


@pytest.mark.parametrize("class_,expected_class,kwargs", SEARCH_IMPLEMENTATIONS)
def test_types(mocker, class_, expected_class, kwargs):
    """Checks all methods of SearchInterface are implemented by this
    backend, and that they have the same signature.

    """
    mocker.patch("swh.search.elasticsearch.Elasticsearch")
    if kwargs:
        concrete_search = get_search(class_, **kwargs)
    else:
        concrete_search = get_search(class_)

    # Create an instance of the protocol (which cannot be instantiated
    # directly, so this creates a subclass, then instantiates it)
    interface = type("_", (SearchInterface,), {})()

    for meth_name in dir(interface):
        if meth_name.startswith("_"):
            continue
        interface_meth = getattr(interface, meth_name)

        missing_methods = []

        try:
            concrete_meth = getattr(concrete_search, meth_name)
        except AttributeError:
            if not getattr(interface_meth, "deprecated_endpoint", False):
                # The backend is missing a (non-deprecated) endpoint
                missing_methods.append(meth_name)
                continue

        expected_signature = inspect.signature(interface_meth)
        actual_signature = inspect.signature(concrete_meth)

        assert expected_signature == actual_signature, meth_name

        assert missing_methods == []



# Copyright (C) 2019-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import Counter
from datetime import datetime, timedelta, timezone
import hashlib
from itertools import permutations

import pytest

from swh.core.api.classes import stream_results


class CommonSearchTest:
    def test_origin_url_unique_word_prefix(self):
        origin_foobar_baz = {
            "url": "http://foobar.baz",
            "visit_types": [],
            "has_visits": False,
        }
        origin_barbaz_qux = {
            "url": "http://barbaz.qux",
            "visit_types": ["git"],
            "has_visits": True,
        }
        origin_qux_quux = {
            "url": "http://qux.quux",
            "visit_types": ["git", "git-checkout"],
            "has_visits": True,
        }
        origins = [origin_foobar_baz, origin_barbaz_qux, origin_qux_quux]

        self.search.origin_update(origins)
        self.search.flush()

        actual_page = self.search.origin_search(url_pattern="foobar")
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin_foobar_baz]

        actual_page = self.search.origin_search(url_pattern="barb")
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin_barbaz_qux]

        # 'bar' is part of 'foobar', but is not the beginning of it
        actual_page = self.search.origin_search(url_pattern="bar")
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin_barbaz_qux]

        actual_page = self.search.origin_search(url_pattern="barbaz")
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin_barbaz_qux]

    def test_origin_url_unique_word_prefix_multiple_results(self):
        origin_foobar_baz = {"url": "http://foobar.baz"}
        origin_barbaz_qux = {"url": "http://barbaz.qux"}
        origin_qux_quux = {"url": "http://qux.quux"}

        self.search.origin_update(
            [origin_foobar_baz, origin_barbaz_qux, origin_qux_quux]
        )
        self.search.flush()

        actual_page = self.search.origin_search(url_pattern="qu")
        assert actual_page.next_page_token is None
        results = [r["url"] for r in actual_page.results]
        expected_results = [o["url"] for o in [origin_qux_quux, origin_barbaz_qux]]
        assert sorted(results) == sorted(expected_results)

        actual_page = self.search.origin_search(url_pattern="qux")
        assert actual_page.next_page_token is None
        results = [r["url"] for r in actual_page.results]
        expected_results = [o["url"] for o in [origin_qux_quux, origin_barbaz_qux]]
        assert sorted(results) == sorted(expected_results)

    def test_origin_url_all_terms(self):
        origin_foo_bar_baz = {
            "url": "http://foo.bar/baz",
            "visit_types": ["git"],
            "has_visits": True,
        }
        origin_foo_bar_foo_bar = {
            "url": "http://foo.bar/foo.bar",
            "visit_types": ["git"],
            "has_visits": True,
        }
        origins = [origin_foo_bar_baz, origin_foo_bar_foo_bar]

        self.search.origin_update(origins)
        self.search.flush()

        # Only results containing all terms should be returned.
        actual_page = self.search.origin_search(url_pattern="foo bar baz")
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin_foo_bar_baz]

    def test_origin_with_visit(self):
        origin_foobar_baz = {
            "url": "http://foobar/baz",
            "visit_types": ["git"],
            "has_visits": True,
        }

        self.search.origin_update(
            [{**o, "has_visits": True} for o in [origin_foobar_baz]]
        )
        self.search.flush()

        actual_page = self.search.origin_search(url_pattern="foobar", with_visit=True)
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin_foobar_baz]

    def test_origin_with_visit_added(self):
        origin_foobar_baz = {
            "url": "http://foobar.baz",
            "visit_types": ["git"],
            "has_visits": False,
        }

        self.search.origin_update([origin_foobar_baz])
        self.search.flush()

        actual_page = self.search.origin_search(url_pattern="foobar", with_visit=True)
        assert actual_page.next_page_token is None
        assert actual_page.results == []

        origin_foobar_baz["has_visits"] = True

        self.search.origin_update([origin_foobar_baz])
        self.search.flush()

        actual_page = self.search.origin_search(url_pattern="foobar", with_visit=True)
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin_foobar_baz]

    def test_origin_no_visit_types_search(self):
        origins = [{"url": "http://foobar.baz"}]

        self.search.origin_update(origins)
        self.search.flush()

        actual_page = self.search.origin_search(url_pattern="http", visit_types=["git"])
        assert actual_page.next_page_token is None
        results = [r["url"] for r in actual_page.results]
        expected_results = []
        assert sorted(results) == sorted(expected_results)

        actual_page = self.search.origin_search(url_pattern="http", visit_types=None)
        assert actual_page.next_page_token is None
        results = [r["url"] for r in actual_page.results]
        expected_results = [origin["url"] for origin in origins]
        assert sorted(results) == sorted(expected_results)

    def test_origin_visit_types_search(self):
        origins = [
            {"url": "http://foobar.baz", "visit_types": ["git"]},
            {"url": "http://barbaz.qux", "visit_types": ["svn"]},
            {"url": "http://qux.quux", "visit_types": ["hg"]},
        ]

        self.search.origin_update(origins)
        self.search.flush()

        for origin in origins:
            actual_page = self.search.origin_search(
                url_pattern="http", visit_types=origin["visit_types"]
            )
            assert actual_page.next_page_token is None
            results = [r["url"] for r in actual_page.results]
            expected_results = [origin["url"]]
            assert sorted(results) == sorted(expected_results)

        actual_page = self.search.origin_search(url_pattern="http", visit_types=None)
        assert actual_page.next_page_token is None
        results = [r["url"] for r in actual_page.results]
        expected_results = [origin["url"] for origin in origins]
        assert sorted(results) == sorted(expected_results)

    def test_origin_visit_types_update_search(self):
        origin_url = "http://foobar.baz"
        self.search.origin_update([{"url": origin_url}])
        self.search.flush()

        def _add_visit_type(visit_type):
            self.search.origin_update(
                [{"url": origin_url, "visit_types": [visit_type]}]
            )
            self.search.flush()

        def _check_visit_types(visit_types_list):
            for visit_types in visit_types_list:
                actual_page = self.search.origin_search(
                    url_pattern="http", visit_types=visit_types
                )
                assert actual_page.next_page_token is None
                results = [r["url"] for r in actual_page.results]
                expected_results = [origin_url]
                assert sorted(results) == sorted(expected_results)

        _add_visit_type("git")
        _check_visit_types([["git"], ["git", "hg"]])

        _add_visit_type("svn")
        _check_visit_types([["git"], ["svn"], ["svn", "git"], ["git", "hg", "svn"]])

        _add_visit_type("hg")
        _check_visit_types(
            [
                ["git"],
                ["svn"],
                ["hg"],
                ["svn", "git"],
                ["hg", "git"],
                ["hg", "svn"],
                ["git", "hg", "svn"],
            ]
        )

    def test_origin_nb_visits_update_search(self):
        origin_url = "http://foobar.baz"
        self.search.origin_update([{"url": origin_url}])
        self.search.flush()

        def _update_nb_visits(nb_visits):
            self.search.origin_update([{"url": origin_url, "nb_visits": nb_visits}])
            self.search.flush()

        def _check_min_nb_visits(min_nb_visits):
            actual_page = self.search.origin_search(
                url_pattern=origin_url,
                min_nb_visits=min_nb_visits,
            )
            assert actual_page.next_page_token is None
            results = [r["url"] for r in actual_page.results]
            expected_results = [origin_url]
            assert sorted(results) == sorted(expected_results)

        _update_nb_visits(2)
        _check_min_nb_visits(2)  # Works for = 2
        _check_min_nb_visits(1)  # Works for < 2

        with pytest.raises(AssertionError):
            _check_min_nb_visits(
                5
            )  # No results for nb_visits >= 5 (should throw error)

        _update_nb_visits(5)
        _check_min_nb_visits(5)  # Works for = 5
        _check_min_nb_visits(3)  # Works for < 5

    def test_origin_last_visit_date_update_search(self):
        origin_url = "http://foobar.baz"
        self.search.origin_update([{"url": origin_url}])
        self.search.flush()

        def _update_last_visit_date(last_visit_date):
            self.search.origin_update(
                [{"url": origin_url, "last_visit_date": last_visit_date}]
            )
            self.search.flush()

        def _check_min_last_visit_date(min_last_visit_date):
            actual_page = self.search.origin_search(
                url_pattern=origin_url,
                min_last_visit_date=min_last_visit_date,
            )
            assert actual_page.next_page_token is None
            results = [r["url"] for r in actual_page.results]
            expected_results = [origin_url]
            assert sorted(results) == sorted(expected_results)

        now = datetime.now(tz=timezone.utc).isoformat()
        now_minus_5_hours = (
            datetime.now(tz=timezone.utc) - timedelta(hours=5)
        ).isoformat()
        now_plus_5_hours = (
            datetime.now(tz=timezone.utc) + timedelta(hours=5)
        ).isoformat()

        _update_last_visit_date(now)

        _check_min_last_visit_date(now)  # Works for =
        _check_min_last_visit_date(now_minus_5_hours)  # Works for <
        with pytest.raises(AssertionError):
            _check_min_last_visit_date(now_plus_5_hours)  # Fails for >

        _update_last_visit_date(now_plus_5_hours)

        _check_min_last_visit_date(now_plus_5_hours)  # Works for =
        _check_min_last_visit_date(now)  # Works for <

    def test_journal_client_origin_visit_status_permutation(self):
        NOW = datetime.now(tz=timezone.utc).isoformat()
        NOW_MINUS_5_HOURS = (
            datetime.now(tz=timezone.utc) - timedelta(hours=5)
        ).isoformat()
        NOW_PLUS_5_HOURS = (
            datetime.now(tz=timezone.utc) + timedelta(hours=5)
        ).isoformat()

        VISIT_STATUSES = [
            {
                "url": "http://foobar.baz",
                "snapshot_id": "SNAPSHOT_1",
                "last_eventful_visit_date": NOW,
            },
            {
                "url": "http://foobar.baz",
                "snapshot_id": "SNAPSHOT_1",
                "last_eventful_visit_date": NOW_MINUS_5_HOURS,
            },
            {
                "url": "http://foobar.baz",
                "snapshot_id": "SNAPSHOT_2",
                "last_eventful_visit_date": NOW_PLUS_5_HOURS,
            },
        ]

        for visit_statuses in permutations(VISIT_STATUSES, len(VISIT_STATUSES)):
            self.search.origin_update(visit_statuses)
            self.search.flush()
            origin_url = "http://foobar.baz"
            actual_page = self.search.origin_search(
                url_pattern=origin_url,
                min_last_eventful_visit_date=NOW_PLUS_5_HOURS,
            )
            assert actual_page.next_page_token is None
            results = [r["url"] for r in actual_page.results]
            expected_results = [origin_url]
            assert sorted(results) == sorted(expected_results)

            self.reset()

    def test_origin_last_eventful_visit_date_update_search(self):
        origin_url = "http://foobar.baz"
        self.search.origin_update([{"url": origin_url}])
        self.search.flush()

        def _update_last_eventful_visit_date(snapshot_id, last_eventful_visit_date):
            self.search.origin_update(
                [
                    {
                        "url": origin_url,
                        "snapshot_id": snapshot_id,
                        "last_eventful_visit_date": last_eventful_visit_date,
                    }
                ]
            )
            self.search.flush()

        def _check_min_last_eventful_visit_date(min_last_eventful_visit_date):
            actual_page = self.search.origin_search(
                url_pattern=origin_url,
                min_last_eventful_visit_date=min_last_eventful_visit_date,
            )
            assert actual_page.next_page_token is None
            results = [r["url"] for r in actual_page.results]
            expected_results = [origin_url]
            assert sorted(results) == sorted(expected_results)

        now = datetime.now(tz=timezone.utc).isoformat()
        now_minus_5_hours = (
            datetime.now(tz=timezone.utc) - timedelta(hours=5)
        ).isoformat()
        now_plus_5_hours = (
            datetime.now(tz=timezone.utc) + timedelta(hours=5)
        ).isoformat()

        snapshot_1 = "SNAPSHOT_1"
        snapshot_2 = "SNAPSHOT_2"

        _update_last_eventful_visit_date(snapshot_1, now)

        _check_min_last_eventful_visit_date(now)  # Works for =
        _check_min_last_eventful_visit_date(now_minus_5_hours)  # Works for <
        with pytest.raises(AssertionError):
            _check_min_last_eventful_visit_date(now_plus_5_hours)  # Fails for >

        _update_last_eventful_visit_date(
            snapshot_1, now_plus_5_hours
        )  # Revisit(not eventful) same origin

        _check_min_last_eventful_visit_date(
            now
        )  # Should remain the same because recent visit wasn't eventful
        with pytest.raises(AssertionError):
            _check_min_last_eventful_visit_date(now_plus_5_hours)

        _update_last_eventful_visit_date(
            snapshot_2, now_plus_5_hours
        )  # Revisit(eventful) same origin
        _check_min_last_eventful_visit_date(now_plus_5_hours)  # Works for =
        _check_min_last_eventful_visit_date(now)  # Works for <

    def _test_origin_last_revision_release_date_update_search(self, date_type):
        origin_url = "http://foobar.baz"
        self.search.origin_update([{"url": origin_url}])
        self.search.flush()

        def _update_last_revision_release_date(date):
            self.search.origin_update(
                [
                    {
                        "url": origin_url,
                        date_type: date,
                    }
                ]
            )
            self.search.flush()

        def _check_min_last_revision_release_date(date):
            actual_page = self.search.origin_search(
                url_pattern=origin_url,
                **{f"min_{date_type}": date},
            )
            assert actual_page.next_page_token is None
            results = [r["url"] for r in actual_page.results]
            expected_results = [origin_url]
            assert sorted(results) == sorted(expected_results)

        now = datetime.now(tz=timezone.utc).isoformat()
        now_minus_5_hours = (
            datetime.now(tz=timezone.utc) - timedelta(hours=5)
        ).isoformat()
        now_plus_5_hours = (
            datetime.now(tz=timezone.utc) + timedelta(hours=5)
        ).isoformat()

        _update_last_revision_release_date(now)

        _check_min_last_revision_release_date(now)
        _check_min_last_revision_release_date(now_minus_5_hours)
        with pytest.raises(AssertionError):
            _check_min_last_revision_release_date(now_plus_5_hours)

        _update_last_revision_release_date(now_plus_5_hours)

        _check_min_last_revision_release_date(now_plus_5_hours)
        _check_min_last_revision_release_date(now)

    def test_origin_last_revision_date_update_search(self):
        self._test_origin_last_revision_release_date_update_search(
            date_type="last_revision_date"
        )

    def test_origin_last_release_date_update_search(self):
        self._test_origin_last_revision_release_date_update_search(
            date_type="last_revision_date"
        )

    def test_origin_instrinsic_metadata_dates_filter_sorting_search(self):
        DATE_0 = "1999-06-28"
        DATE_1 = "2001-02-13"
        DATE_2 = "2005-10-02"

        ORIGINS = [
            {
                "url": "http://foobar.0.com",
                "jsonld": {
                    "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                    "dateCreated": DATE_0,
                    "dateModified": DATE_1,
                    "datePublished": DATE_2,
                },
            },
            {
                "url": "http://foobar.1.com",
                "jsonld": {
                    "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                    "dateCreated": DATE_1,
                    "dateModified": DATE_2,
                    "datePublished": DATE_2,
                },
            },
            {
                "url": "http://foobar.2.com",
                "jsonld": {
                    "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                    "dateCreated": DATE_2,
                    "dateModified": DATE_2,
                    "datePublished": DATE_2,
                },
            },
        ]
        self.search.origin_update(ORIGINS)
        self.search.flush()

        def _check_results(origin_indices, sort_results=True, **kwargs):
            page = self.search.origin_search(url_pattern="foobar", **kwargs)
            results = [r["url"] for r in page.results]
            if sort_results:
                assert sorted(results) == sorted(
                    [ORIGINS[index]["url"] for index in origin_indices]
                )
            else:
                assert results == [ORIGINS[index]["url"] for index in origin_indices]

        _check_results(min_date_created=DATE_0, origin_indices=[0, 1, 2])
        _check_results(min_date_created=DATE_1, origin_indices=[1, 2])
        _check_results(min_date_created=DATE_2, origin_indices=[2])

        _check_results(min_date_modified=DATE_0, origin_indices=[0, 1, 2])
        _check_results(min_date_modified=DATE_1, origin_indices=[0, 1, 2])
        _check_results(min_date_modified=DATE_2, origin_indices=[1, 2])

        _check_results(min_date_published=DATE_0, origin_indices=[0, 1, 2])
        _check_results(min_date_published=DATE_1, origin_indices=[0, 1, 2])
        _check_results(min_date_published=DATE_2, origin_indices=[0, 1, 2])

        # Sorting
        _check_results(
            sort_by=["-date_created"], origin_indices=[2, 1, 0], sort_results=False
        )
        _check_results(
            sort_by=["date_created"], origin_indices=[0, 1, 2], sort_results=False
        )

    def test_origin_instrinsic_metadata_dates_processing(self):
        DATE_0 = "foo"  # will be discarded
        DATE_1 = "2001-2-13"  # will be formatted to 2001-02-13
        DATE_2 = "2005-10-2"  # will be formatted to 2005-10-02

        ORIGINS = [
            {
                "url": "http://foobar.0.com",
                "jsonld": {
                    "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                    "dateCreated": DATE_0,
                    "dateModified": DATE_1,
                    "datePublished": DATE_2,
                },
            },
            {
                "url": "http://foobar.1.com",
                "jsonld": {
                    "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                    "dateCreated": DATE_1,
                    "dateModified": DATE_2,
                    "datePublished": DATE_2,
                },
            },
            {
                "url": "http://foobar.2.com",
                "jsonld": {
                    "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                    "dateCreated": DATE_2,
                    "dateModified": DATE_2,
                    "datePublished": DATE_2,
                },
            },
        ]
        self.search.origin_update(ORIGINS)
        self.search.flush()

        # check origins have been successfully processed
        page = self.search.origin_search(url_pattern="foobar")
        assert {r["url"] for r in page.results} == {
            "http://foobar.0.com",
            "http://foobar.2.com",
            "http://foobar.1.com",
        }

    def test_origin_keywords_search(self):
        ORIGINS = [
            {
                "url": "http://foobar.1.com",
                "jsonld": {
                    "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                    "description": "Django is a backend framework for applications",
                    "keywords": "django,backend,server,web,framework",
                },
            },
            {
                "url": "http://foobar.2.com",
                "jsonld": {
                    "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                    "description": "Native Android applications are fast",
                    "keywords": "android,mobile,ui",
                },
            },
            {
                "url": "http://foobar.3.com",
                "jsonld": {
                    "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                    "description": "React framework helps you build web applications",
                    "keywords": "react,web,ui",
                },
            },
        ]
        self.search.origin_update(ORIGINS)
        self.search.flush()

        def _check_results(keywords, origin_indices, sorting=False):
            page = self.search.origin_search(url_pattern="foobar", keywords=keywords)
            results = [r["url"] for r in page.results]
            if sorting:
                assert sorted(results) == sorted(
                    [ORIGINS[index]["url"] for index in origin_indices]
                )
            else:
                assert results == [ORIGINS[index]["url"] for index in origin_indices]

        _check_results(["build"], [2])

        _check_results(["web"], [2, 0])
        _check_results(["ui"], [1, 2])

        # Following tests ensure that boosts work properly

        # Baseline: "applications" is common in all origin descriptions
        _check_results(["applications"], [1, 0, 2], True)

        # ORIGINS[0] has 'framework' in: keyword + description
        # ORIGINS[2] has 'framework' in: description
        # ORIGINS[1] has 'framework' in: None
        _check_results(["framework", "applications"], [0, 2, 1])

        # ORIGINS[1] has 'ui' in: keyword
        # ORIGINS[1] has 'ui' in: keyword
        # ORIGINS[0] has 'ui' in: None
        _check_results(["applications", "ui"], [1, 2, 0])

        # ORIGINS[2] has 'web' in: keyword + description
        # ORIGINS[0] has 'web' in: keyword
        # ORIGINS[1] has 'web' in: None
        _check_results(["web", "applications"], [2, 0, 1])

    def test_origin_sort_by_search(self):
        now = datetime.now(tz=timezone.utc).isoformat()
        now_minus_5_hours = (
            datetime.now(tz=timezone.utc) - timedelta(hours=5)
        ).isoformat()
        now_plus_5_hours = (
            datetime.now(tz=timezone.utc) + timedelta(hours=5)
        ).isoformat()

        ORIGINS = [
            {
                "url": "http://foobar.1.com",
                "nb_visits": 1,
                "last_visit_date": now_minus_5_hours,
            },
            {
                "url": "http://foobar.2.com",
                "nb_visits": 2,
                "last_visit_date": now,
            },
            {
                "url": "http://foobar.3.com",
                "nb_visits": 3,
                "last_visit_date": now_plus_5_hours,
            },
        ]
        self.search.origin_update(ORIGINS)
        self.search.flush()

        def _check_results(sort_by, origins):
            page = self.search.origin_search(url_pattern="foobar", sort_by=sort_by)
            results = [r["url"] for r in page.results]
            assert results == [origin["url"] for origin in origins]

        _check_results(["nb_visits"], ORIGINS)
        _check_results(["-nb_visits"], ORIGINS[::-1])

        _check_results(["last_visit_date"], ORIGINS)
        _check_results(["-last_visit_date"], ORIGINS[::-1])

        _check_results(["nb_visits", "-last_visit_date"], ORIGINS)
        _check_results(["-last_visit_date", "nb_visits"], ORIGINS[::-1])

    def test_origin_instrinsic_metadata_license_search(self):
        ORIGINS = [
            {
                "url": "http://foobar.1.com",
                "jsonld": {
                    "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                    "description": "foo bar",
                    "license": "https://spdx.org/licenses/MIT",
                },
            },
            {
                "url": "http://foobar.2.com",
                "jsonld": {
                    "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                    "description": "foo bar",
                    "license": "BSD-3-Clause",
                },
            },
        ]
        self.search.origin_update(ORIGINS)
        self.search.flush()

        def _check_results(licenses, origin_indices):
            page = self.search.origin_search(url_pattern="foobar", licenses=licenses)
            results = [r["url"] for r in page.results]
            assert sorted(results) == sorted(
                [ORIGINS[i]["url"] for i in origin_indices]
            )

        _check_results(["MIT"], [0])
        _check_results(["bsd"], [1])
        _check_results(["mit", "3-Clause"], [0, 1])

    def test_origin_instrinsic_metadata_programming_language_search(self):
        ORIGINS = [
            {
                "url": "http://foobar.1.com",
                "jsonld": {
                    "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                    "description": "foo bar",
                    "programmingLanguage": "python",
                },
            },
            {
                "url": "http://foobar.2.com",
                "jsonld": {
                    "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                    "description": "foo bar",
                    "programmingLanguage": "javascript",
                },
            },
        ]
        self.search.origin_update(ORIGINS)
        self.search.flush()

        def _check_results(programming_languages, origin_indices):
            page = self.search.origin_search(
                url_pattern="foobar", programming_languages=programming_languages
            )
            results = [r["url"] for r in page.results]
            assert sorted(results) == sorted(
                [ORIGINS[i]["url"] for i in origin_indices]
            )

        _check_results(["python"], [0])
        _check_results(["javascript"], [1])
        _check_results(["python", "javascript"], [0, 1])

    def test_origin_instrinsic_metadata_multiple_field_search(self):
        ORIGINS = [
            {
                "url": "http://foobar.1.com",
                "jsonld": {
                    "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                    "description": "foo bar 1",
                    "programmingLanguage": "python",
                    "license": "https://spdx.org/licenses/MIT",
                },
            },
            {
                "url": "http://foobar.2.com",
                "jsonld": {
                    "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                    "description": "foo bar 2",
                    "programmingLanguage": ["javascript", "html", "css"],
                    "license": [
                        "https://spdx.org/licenses/CC-BY-1.0",
                        "https://spdx.org/licenses/Apache-1.0",
                    ],
                },
            },
            {
                "url": "http://foobar.3.com",
                "jsonld": {
                    "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                    "description": "foo bar 3",
                    "programmingLanguage": ["Cpp", "c"],
                    "license": "https://spdx.org/licenses/LGPL-2.0-only",
                },
            },
        ]
        self.search.origin_update(ORIGINS)
        self.search.flush()

        def _check_result(programming_languages, licenses, origin_indices):
            page = self.search.origin_search(
                url_pattern="foobar",
                programming_languages=programming_languages,
                licenses=licenses,
            )
            results = [r["url"] for r in page.results]
            assert sorted(results) == sorted(
                [ORIGINS[i]["url"] for i in origin_indices]
            )

        _check_result(["javascript"], ["CC"], [1])
        _check_result(["css"], ["CC"], [1])
        _check_result(["css"], ["CC", "apache"], [1])

        _check_result(["python", "javascript"], ["MIT"], [0])

        _check_result(["c", "python"], ["LGPL", "mit"], [2, 0])

    def test_origin_update_with_no_visit_types(self):
        """
        Update an origin with visit types first then with no visit types,
        check origin can still be searched with visit types afterwards.
        """
        origin_url = "http://foobar.baz"
        self.search.origin_update([{"url": origin_url, "visit_types": ["git"]}])
        self.search.flush()

        self.search.origin_update([{"url": origin_url}])
        self.search.flush()

        actual_page = self.search.origin_search(url_pattern="http", visit_types=["git"])
        assert actual_page.next_page_token is None
        results = [r["url"] for r in actual_page.results]
        expected_results = [origin_url]
        assert results == expected_results

    def test_origin_jsonld_description(self):
        origin1_nothin = {
            "url": "http://origin1",
            "visit_types": ["git"],
            "has_visits": True,
        }
        origin2_foobar = {
            "url": "http://origin2",
            "visit_types": ["git"],
            "has_visits": True,
        }
        origin3_barbaz = {
            "url": "http://origin3",
            "visit_types": ["git"],
            "has_visits": True,
        }

        self.search.origin_update(
            [
                {
                    **origin1_nothin,
                    "jsonld": {},
                },
                {
                    **origin2_foobar,
                    "jsonld": {
                        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                        "description": "foo bar",
                    },
                },
                {
                    **origin3_barbaz,
                    "jsonld": {
                        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                        "description": "bar baz",
                    },
                },
            ]
        )
        self.search.flush()

        actual_page = self.search.origin_search(metadata_pattern="foo")
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin2_foobar]

        actual_page = self.search.origin_search(metadata_pattern="foo bar")
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin2_foobar]

        actual_page = self.search.origin_search(metadata_pattern="bar baz")
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin3_barbaz]

    def test_origin_jsonld_all_terms(self):
        origin1_foobarfoobar = {
            "url": "http://origin1",
            "visit_types": ["git"],
            "has_visits": True,
        }
        origin3_foobarbaz = {
            "url": "http://origin2",
            "visit_types": ["git"],
            "has_visits": True,
        }

        self.search.origin_update(
            [
                {
                    **origin1_foobarfoobar,
                    "jsonld": {
                        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                        "description": "foo bar foo bar",
                    },
                },
                {
                    **origin3_foobarbaz,
                    "jsonld": {
                        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                        "description": "foo bar baz",
                    },
                },
            ]
        )
        self.search.flush()

        actual_page = self.search.origin_search(metadata_pattern="foo bar baz")
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin3_foobarbaz]

    def test_origin_jsonld_long_description(self):
        """Checks ElasticSearch does not try to store large values untokenize,
        which would be inefficient and crash it with:

        Document contains at least one immense term in field="jsonld.http://schema.org/description.@value" (whose UTF8 encoding is longer than the max length 32766), all of which were skipped.
        """  # noqa
        origin1 = {
            "url": "http://origin1",
            "visit_types": ["git"],
            "has_visits": True,
        }

        self.search.origin_update(
            [
                {
                    **origin1,
                    "jsonld": {
                        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                        "description": " ".join(f"foo{i}" for i in range(100000)),
                    },
                },
            ]
        )
        self.search.flush()

        actual_page = self.search.origin_search(metadata_pattern="foo42")
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin1]

    def test_origin_jsonld_matches_cross_fields(self):
        """Checks the backend finds results even if the two words in the query are
        each in a different field."""
        origin1 = {
            "url": "http://origin1",
            "visit_types": ["git"],
            "has_visits": True,
        }

        self.search.origin_update(
            [
                {
                    **origin1,
                    "jsonld": {
                        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                        "description": "foo bar",
                        "author": "John Doe",
                    },
                },
            ]
        )
        self.search.flush()

        actual_page = self.search.origin_search(metadata_pattern="foo John")
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin1]

    def test_origin_jsonld_nested(self):
        origin1_nothin = {
            "url": "http://origin1",
            "visit_types": ["git"],
            "has_visits": True,
        }
        origin2_foobar = {
            "url": "http://origin2",
            "visit_types": ["git"],
            "has_visits": True,
        }
        origin3_barbaz = {
            "url": "http://origin3",
            "visit_types": ["git"],
            "has_visits": True,
        }

        self.search.origin_update(
            [
                {
                    **origin1_nothin,
                    "jsonld": {},
                },
                {
                    **origin2_foobar,
                    "jsonld": {
                        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                        "keywords": ["foo", "bar"],
                    },
                },
                {
                    **origin3_barbaz,
                    "jsonld": {
                        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                        "keywords": ["bar", "baz"],
                    },
                },
            ]
        )
        self.search.flush()

        actual_page = self.search.origin_search(metadata_pattern="foo")
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin2_foobar]

        actual_page = self.search.origin_search(metadata_pattern="foo bar")
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin2_foobar]

        actual_page = self.search.origin_search(metadata_pattern="bar baz")
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin3_barbaz]

    def test_origin_jsonld_inconsistent_type(self):
        """Checks the same field can have a concrete value, an object, or an array
        in different documents."""
        origin1_foobar = {
            "url": "http://origin1",
            "visit_types": ["git"],
            "has_visits": True,
        }
        origin2_barbaz = {
            "url": "http://origin2",
            "visit_types": ["git"],
            "has_visits": True,
        }
        origin3_bazqux = {
            "url": "http://origin3",
            "visit_types": ["git"],
            "has_visits": True,
        }

        self.search.origin_update(
            [
                {
                    **origin1_foobar,
                    "jsonld": {
                        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                        "author": {
                            "familyName": "Foo",
                            "givenName": "Bar",
                        },
                    },
                },
            ]
        )
        self.search.flush()
        self.search.origin_update(
            [
                {
                    **origin2_barbaz,
                    "jsonld": {
                        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                        "author": "Bar Baz",
                    },
                },
                {
                    **origin3_bazqux,
                    "jsonld": {
                        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                        "author": ["Baz", "Qux"],
                    },
                },
            ]
        )
        self.search.flush()

        actual_page = self.search.origin_search(metadata_pattern="bar")
        assert actual_page.next_page_token is None
        results = [r["url"] for r in actual_page.results]
        expected_results = [o["url"] for o in [origin2_barbaz, origin1_foobar]]
        assert sorted(results) == sorted(expected_results)

        actual_page = self.search.origin_search(metadata_pattern="baz")
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin2_barbaz, origin3_bazqux]

        actual_page = self.search.origin_search(metadata_pattern="foo")
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin1_foobar]

        actual_page = self.search.origin_search(metadata_pattern="bar baz")
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin2_barbaz]

        actual_page = self.search.origin_search(metadata_pattern="qux")
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin3_bazqux]

        actual_page = self.search.origin_search(metadata_pattern="baz qux")
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin3_bazqux]

        actual_page = self.search.origin_search(metadata_pattern="foo bar")
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin1_foobar]

    def test_origin_jsonld_string_mapping(self):
        """Checks inserting a date-like in a field does not update the mapping to
        require every document uses a date in that field; or that search queries
        use a date either.
        Likewise for numeric and boolean fields."""
        origin1 = {
            "url": "http://origin1",
            "visit_types": ["git"],
            "has_visits": True,
        }
        origin2 = {
            "url": "http://origin2",
            "visit_types": ["git"],
            "has_visits": True,
        }

        self.search.origin_update(
            [
                {
                    **origin1,
                    "jsonld": {
                        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                        "dateCreated": "2021-02-18T10:16:52",
                        "version": 1.0,
                        "softwareVersion": "1.0",
                        "isAccessibleForFree": True,
                        "copyrightYear": 2022,
                    },
                }
            ]
        )
        self.search.flush()
        self.search.origin_update(
            [
                {
                    **origin2,
                    "jsonld": {
                        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                        "dateCreated": "a long time ago",
                        "address": "in a galaxy far, far away",
                        "version": "a new hope",
                        "softwareVersion": "a new hope",
                        "isAccessibleForFree": "it depends",
                        "copyrightYear": "foo bar",
                    },
                },
            ]
        )
        self.search.flush()

        actual_page = self.search.origin_search(metadata_pattern="1.0")
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin1]

        actual_page = self.search.origin_search(metadata_pattern="long")
        assert actual_page.next_page_token is None
        assert (
            actual_page.results == []
        )  # "%Y-%m-%d" not followed, so value is rejected

        actual_page = self.search.origin_search(metadata_pattern="true")
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin1]

        actual_page = self.search.origin_search(metadata_pattern="it depends")
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin2]

    def test_origin_jsonld_rank_forks(self):
        """Checks forks rank lower than "original" repositories."""

        # mark http://foobar.2.com as a fork of http://foobar.1.com
        origins = [
            {
                "url": "http://foobar.1.com",
            },
            {
                "url": "http://foobar.2.com",
                "jsonld": {
                    "https://forgefed.org/ns#forkedFrom": [
                        {"@id": "http://foobar.1.com"},
                    ],
                },
            },
            {
                "url": "http://foobar.3.org",
                "jsonld": {
                    # Unrelated JSON-LD fields should have no impact
                    "http://schema.org/dateCreated": "2001-02-02",
                },
            },
        ]
        self.search.origin_update(origins)
        self.search.flush()

        results = [
            r["url"] for r in self.search.origin_search(url_pattern="foobar").results
        ]
        assert set(results[0:2]) == {
            "http://foobar.1.com",
            "http://foobar.3.org",
        }, results
        assert results[2] == "http://foobar.2.com", results

        # Boost forks instead of penalizing them
        results = [
            r["url"]
            for r in self.search.origin_search(
                url_pattern="foobar", fork_weight=2.0
            ).results
        ]
        assert results[0] == "http://foobar.2.com", results
        assert set(results[1:]) == {
            "http://foobar.1.com",
            "http://foobar.3.org",
        }, results

    def test_origin_jsonld_update(self):
        origin = {
            "url": "http://origin1",
            "visit_types": ["git"],
            "has_visits": True,
        }
        origin_data = {
            **origin,
            "jsonld": {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "author": "John Doe",
            },
        }

        self.search.origin_update([origin_data])
        self.search.flush()

        actual_page = self.search.origin_search(metadata_pattern="John")
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin]

        origin_data["jsonld"]["author"] = "Jane Doe"

        self.search.origin_update([origin_data])
        self.search.flush()

        actual_page = self.search.origin_search(metadata_pattern="Jane")
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin]

    # TODO: add more tests with more codemeta terms

    # TODO: add more tests with edge cases

    @pytest.mark.parametrize("limit", list(range(1, 5)))
    def test_origin_url_paging(self, limit):
        origin1_foo = {
            "url": "http://origin1/foo",
            "visit_types": ["git"],
            "has_visits": True,
        }
        origin2_foobar = {
            "url": "http://origin2/foo/bar",
            "visit_types": ["git"],
            "has_visits": True,
        }
        origin3_foobarbaz = {
            "url": "http://origin3/foo/bar/baz",
            "visit_types": ["git"],
            "has_visits": True,
        }

        self.reset()
        self.search.origin_update([origin1_foo, origin2_foobar, origin3_foobarbaz])
        self.search.flush()

        results = stream_results(
            self.search.origin_search, url_pattern="foo bar baz", limit=limit
        )
        results = [res["url"] for res in results]
        expected_results = [o["url"] for o in [origin3_foobarbaz]]
        assert sorted(results[0 : len(expected_results)]) == sorted(expected_results)

        results = stream_results(
            self.search.origin_search, url_pattern="foo bar", limit=limit
        )
        results = [res["url"] for res in results]
        expected_results = [o["url"] for o in [origin2_foobar, origin3_foobarbaz]]
        assert sorted(results[0 : len(expected_results)]) == sorted(expected_results)

        results = stream_results(
            self.search.origin_search, url_pattern="foo", limit=limit
        )
        results = [res["url"] for res in results]
        expected_results = [
            o["url"] for o in [origin1_foo, origin2_foobar, origin3_foobarbaz]
        ]
        assert sorted(results[0 : len(expected_results)]) == sorted(expected_results)

    @pytest.mark.parametrize("limit", list(range(1, 5)))
    def test_origin_jsonld_paging(self, limit):
        origin1_foo = {
            "url": "http://origin1",
            "visit_types": ["git"],
            "has_visits": True,
        }
        origin2_foobar = {
            "url": "http://origin2",
            "visit_types": ["git"],
            "has_visits": True,
        }
        origin3_foobarbaz = {
            "url": "http://origin3",
            "visit_types": ["git"],
            "has_visits": True,
        }

        self.reset()
        self.search.origin_update(
            [
                {
                    **origin1_foo,
                    "jsonld": {
                        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                        "keywords": ["foo"],
                    },
                },
                {
                    **origin2_foobar,
                    "jsonld": {
                        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                        "keywords": ["foo", "bar"],
                    },
                },
                {
                    **origin3_foobarbaz,
                    "jsonld": {
                        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                        "keywords": ["foo", "bar", "baz"],
                    },
                },
            ]
        )
        self.search.flush()

        results = stream_results(
            self.search.origin_search, metadata_pattern="foo bar baz", limit=limit
        )
        assert list(results) == [origin3_foobarbaz]

        results = stream_results(
            self.search.origin_search, metadata_pattern="foo bar", limit=limit
        )
        assert list(results) == [origin2_foobar, origin3_foobarbaz]

        results = stream_results(
            self.search.origin_search, metadata_pattern="foo", limit=limit
        )
        assert list(results) == [origin1_foo, origin2_foobar, origin3_foobarbaz]

    def test_search_blocklisted_results(self):
        origin1 = {
            "url": "http://origin1",
            "visit_types": ["git"],
            "has_visits": True,
        }
        origin2 = {
            "url": "http://origin2",
            "visit_types": ["git"],
            "has_visits": True,
            "blocklisted": True,
        }

        self.search.origin_update([origin1, origin2])
        self.search.flush()

        actual_page = self.search.origin_search(url_pattern="origin")
        assert actual_page.next_page_token is None
        assert actual_page.results == [origin1]

    def test_search_blocklisted_update(self):
        origin1 = {
            "url": "http://origin1",
            "visit_types": ["git"],
            "has_visits": True,
        }
        self.search.origin_update([origin1])
        self.search.flush()

        result_page = self.search.origin_search(url_pattern="origin")
        assert result_page.next_page_token is None
        assert result_page.results == [origin1]

        self.search.origin_update([{**origin1, "blocklisted": True}])
        self.search.flush()

        result_page = self.search.origin_search(url_pattern="origin")
        assert result_page.next_page_token is None
        assert result_page.results == []

        self.search.origin_update(
            [{**origin1, "has_visits": True, "visit_types": ["git"]}]
        )
        self.search.flush()

        result_page = self.search.origin_search(url_pattern="origin")
        assert result_page.next_page_token is None
        assert result_page.results == []

    def test_search_filter_keyword_in_filter(self):
        origin1 = {
            "url": "foo language in ['foo baz'] bar",
            "visit_types": ["git"],
            "has_visits": True,
        }
        self.search.origin_update([origin1])
        self.search.flush()

        result_page = self.search.origin_search(url_pattern="language in ['foo bar']")
        assert result_page.next_page_token is None
        assert result_page.results == [origin1]

        result_page = self.search.origin_search(url_pattern="baaz")
        assert result_page.next_page_token is None
        assert result_page.results == []

    def test_origin_get(self):
        """Checks the same field can have a concrete value, an object, or an array
        in different documents."""
        origin1 = {"url": "http://origin1", "visit_types": ["git"], "has_visits": True}
        origin2 = {"url": "http://origin2", "visit_types": ["git"], "has_visits": True}
        origin3 = {"url": "http://origin3", "visit_types": ["git"], "has_visits": True}
        origins = [
            {
                **origin1,
                "jsonld": {
                    "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                    "author": {
                        "familyName": "Foo",
                        "givenName": "Bar",
                    },
                },
            },
            {
                **origin2,
                "jsonld": {
                    "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                    "author": "Bar Baz",
                },
            },
            {
                **origin3,
                "jsonld": {
                    "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                    "author": ["Baz", "Qux"],
                },
            },
        ]

        expanded_origins = [
            {
                **origin1,
                "sha1": hashlib.sha1(origin1["url"].encode()).hexdigest(),
                "jsonld": [
                    {
                        "http://schema.org/author": [
                            {
                                "@list": [
                                    {
                                        "http://schema.org/familyName": [
                                            {"@value": "Foo"}
                                        ],
                                        "http://schema.org/givenName": [
                                            {"@value": "Bar"}
                                        ],
                                    }
                                ]
                            }
                        ],
                    }
                ],
            },
            {
                **origin2,
                "sha1": hashlib.sha1(origin2["url"].encode()).hexdigest(),
                "jsonld": [
                    {
                        "http://schema.org/author": [
                            {"@list": [{"@value": "Bar Baz"}]}
                        ],
                    }
                ],
            },
            {
                **origin3,
                "sha1": hashlib.sha1(origin3["url"].encode()).hexdigest(),
                "jsonld": [
                    {
                        "http://schema.org/author": [
                            {"@list": [{"@value": "Baz"}, {"@value": "Qux"}]}
                        ],
                    }
                ],
            },
        ]

        self.search.origin_update(origins)
        self.search.flush()

        assert self.search.origin_get(origin1["url"]) == expanded_origins[0]
        assert self.search.origin_get(origin2["url"]) == expanded_origins[1]
        assert self.search.origin_get(origin3["url"]) == expanded_origins[2]
        assert self.search.origin_get("http://origin4") is None

    def test_origin_delete(self):
        origin1 = {"url": "http://one", "visit_types": ["git"], "has_visits": True}
        origin2 = {"url": "http://two", "visit_types": ["git"], "has_visits": True}
        self.search.origin_update([origin1, origin2])
        self.search.flush()

        assert self.search.origin_get("http://one")["url"] == "http://one"
        assert len(self.search.origin_search(url_pattern="one", limit=1).results) == 1
        deleted = self.search.origin_delete("http://one")
        assert deleted, "origin not found"
        self.search.flush()

        assert self.search.origin_get("http://one") is None
        assert len(self.search.origin_search(url_pattern="one", limit=1).results) == 0
        assert self.search.origin_get("http://two")["url"] == "http://two"

        # Ensure idempotency
        deleted = self.search.origin_delete("http://one")
        self.search.flush()
        assert not deleted, "origin deleted twice‽"
        assert self.search.origin_get("http://one") is None

    def test_visit_types_count(self):
        assert self.search.visit_types_count() == Counter()

        origins = [
            {"url": "http://foobar.baz", "visit_types": ["git"], "blocklisted": True}
        ]

        for idx, visit_type in enumerate(["git", "hg", "svn"]):
            for i in range(idx + 1):
                origins.append(
                    {
                        "url": f"http://{visit_type}.foobar.baz.{i}",
                        "visit_types": [visit_type],
                    }
                )
        self.search.origin_update(origins)
        self.search.flush()

        assert self.search.visit_types_count() == Counter(git=1, hg=2, svn=3)

    def test_origin_search_empty_url_pattern(self):
        origins = [
            {"url": "http://foobar.baz", "visit_types": ["git"]},
            {"url": "http://barbaz.qux", "visit_types": ["svn"]},
            {"url": "http://qux.quux", "visit_types": ["hg"]},
        ]

        self.search.origin_update(origins)
        self.search.flush()

        # should match all origins
        actual_page = self.search.origin_search(url_pattern="")
        assert actual_page.next_page_token is None
        results = [r["url"] for r in actual_page.results]
        expected_results = [origin["url"] for origin in origins]
        assert sorted(results) == sorted(expected_results)

        # should match all origins with visit type
        for origin in origins:
            actual_page = self.search.origin_search(
                url_pattern="", visit_types=origin["visit_types"]
            )
            assert actual_page.next_page_token is None
            results = [r["url"] for r in actual_page.results]
            expected_results = [origin["url"]]
            assert results == expected_results

    def test_origin_search_total_results(self):
        limit = 10
        nb_git_origins = 100
        origins = [
            {
                "url": f"http://git.example.org/user/project{i:03d}",
                "visit_types": ["git"],
            }
            for i in range(nb_git_origins)
        ]
        nb_svn_origins = 70
        origins += [
            {
                "url": f"http://svn.example.org/user/project{i:03d}",
                "visit_types": ["svn"],
            }
            for i in range(nb_svn_origins)
        ]

        self.search.origin_update(origins)
        self.search.flush()

        page = self.search.origin_search(
            url_pattern="http://git.example.org", limit=limit
        )
        assert len(page.results) == limit
        assert page.total_results == nb_git_origins

        page = self.search.origin_search(
            url_pattern="http://svn.example.org", limit=limit
        )
        assert len(page.results) == limit
        assert page.total_results == nb_svn_origins



# Copyright (C) 2019-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
from pathlib import Path
import socket
import subprocess
import textwrap
import time

import elasticsearch
import pytest

from swh.search import get_search

logger = logging.getLogger(__name__)


def free_port():
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("127.0.0.1", 0))
    port = sock.getsockname()[1]
    sock.close()
    return port


CONFIG_TEMPLATES = {
    "elasticsearch.yml": """\
        action.auto_create_index: false
        action.destructive_requires_name: false
        cluster.routing.allocation.disk.threshold_enabled: false
        discovery.type: single-node
        http.port: {http_port}
        network.host: 127.0.0.1
        node.name: node-1
        path.data: {data}
        path.logs: {logs}
        stack.templates.enabled: false
        transport.port: {transport_port}
        xpack.ml.enabled: false
        xpack.security.enabled: false
    """,
    "jvm.options": """\
        -XX:+UseG1GC
        -Djava.io.tmpdir={data}
        20-:--add-modules=jdk.incubator.vector
        -XX:+HeapDumpOnOutOfMemoryError
        -XX:+ExitOnOutOfMemoryError
        -XX:HeapDumpPath={logs}
        -XX:ErrorFile={logs}/hs_err_pid%p.log
        -Xlog:gc*,gc+age=trace,safepoint:file={logs}/gc.log:utctime,level,pid,tags:filecount=32,filesize=64m
    """,
    "log4j2.properties": """\
        status = error

        appender.console.type = Console
        appender.console.name = console
        appender.console.layout.type = PatternLayout
        appender.console.layout.pattern = [%d{{ISO8601}}][%-5p][%-25c{{1.}}] [%node_name]%marker %m%consoleException%n

        rootLogger.level = info
        rootLogger.appenderRef.console.ref = console
    """,  # noqa
}

ES_HOME = "/usr/share/elasticsearch"


def _get_es_version():
    """Parse /usr/share/elasticsearch/lib/elasticsearch-X.Y.Z.jar into X"""
    jars = list(Path(f"{ES_HOME}/lib").glob("elasticsearch-[0-9]*.jar"))
    if len(jars) != 1:
        raise ValueError(f"Found unexpected contents in {ES_HOME}/lib: {jars}")

    jarfile = jars[0]
    major = jarfile.stem.split("-")[-1].split(".")[0]
    return int(major)


def _run_elasticsearch(
    conf_dir, data_dir, logs_dir, http_port, transport_port, libffi_tmpdir
):
    for filename, template in CONFIG_TEMPLATES.items():
        with open(f"{conf_dir}/{filename}", "w") as fd:
            fd.write(
                textwrap.dedent(template).format(
                    data=data_dir,
                    logs=logs_dir,
                    http_port=http_port,
                    transport_port=transport_port,
                )
            )

    major = _get_es_version()
    if major >= 8:
        cmd = [
            f"{ES_HOME}/jdk/bin/java",
            "-Dcli.name=server",
            f"-Dcli.script={ES_HOME}/bin/elasticsearch",
            "-Dcli.libs=lib/tools/server-cli",
            f"-Des.path.home={ES_HOME}",
            f"-Des.path.conf={conf_dir}",
            "-Des.distribution.type=deb",
            "-Des.bundled_jdk=true",
            "-cp",
            f"{ES_HOME}/lib/*:{ES_HOME}/lib/cli-launcher/*",
            "org.elasticsearch.launcher.CliToolLauncher",
        ]
    else:
        cmd = [
            f"{ES_HOME}/jdk/bin/java",
            f"-Des.path.home={ES_HOME}",
            f"-Des.path.conf={conf_dir}",
            "-Des.bundled_jdk=true",
            "-Dlog4j2.disable.jmx=true",
            # fix UnsupportedOperationException related to deprecated Security Manager
            # when using JDK >= 18
            "-Djava.security.manager=allow",
            "-cp",
            f"{ES_HOME}/lib/*",
            "org.elasticsearch.bootstrap.Elasticsearch",
        ]

    process = subprocess.Popen(cmd, cwd=ES_HOME, env={"LIBFFI_TMPDIR": libffi_tmpdir})

    for _ in range(3000):
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect(("127.0.0.1", http_port))
        except ConnectionRefusedError:
            if process.poll() is not None:

                # process has exited
                break

            time.sleep(0.1)
        else:
            sock.close()
            break

    host = f"127.0.0.1:{http_port}"
    client = elasticsearch.Elasticsearch([host])
    assert client.ping()

    return process, major


@pytest.fixture(scope="session")
def elasticsearch_session(tmpdir_factory):
    tmpdir = tmpdir_factory.mktemp("elasticsearch")
    es_conf = tmpdir.mkdir("conf")

    http_port = free_port()
    transport_port = free_port()

    proc, major = _run_elasticsearch(
        conf_dir=str(es_conf),
        data_dir=str(tmpdir.mkdir("data")),
        logs_dir=str(tmpdir.mkdir("logs")),
        http_port=http_port,
        transport_port=transport_port,
        libffi_tmpdir=str(tmpdir.mkdir("libffi")),
    )

    yield "127.0.0.1:{}".format(http_port)

    # Check ES didn't stop
    assert proc.returncode is None, proc.returncode

    proc.terminate()
    if major < 8:
        proc.kill()
    proc.wait()


@pytest.fixture(scope="class")
def elasticsearch_host(elasticsearch_session):
    yield elasticsearch_session


@pytest.fixture
def swh_search(elasticsearch_host):
    """Instantiate a search client, initialize the elasticsearch instance,
    and returns it

    """
    logger.debug("swh_search: elasticsearch_host: %s", elasticsearch_host)
    search = get_search(
        "elasticsearch",
        hosts=[elasticsearch_host],
        indexes={
            "origin": {
                "index": "test",
                "read_alias": "test-read",
                "write_alias": "test-write",
            }
        },
    )
    search.deinitialize()  # To reset internal state from previous runs
    search.initialize()  # install required index
    yield search



# Copyright (C) 2025 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

pytest_plugins = ["swh.journal.pytest_plugin"]



# Copyright (C) 2018-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import sys
from typing import Dict, Optional

from swh.model.model import SnapshotTargetType
from swh.storage.algos.snapshot import snapshot_get_all_branches
from swh.storage.interface import StorageInterface

EXPECTED_MESSAGE_TYPES = {
    "origin",
    "origin_visit_status",
    "origin_intrinsic_metadata",
    "origin_extrinsic_metadata",
}


def fetch_last_revision_release_date(
    snapshot_id: bytes, storage: StorageInterface
) -> Dict[str, str]:
    if "pytest" not in sys.modules:
        # FIXME: This function is too slow to be reasonably used in the journal-client
        # (at least the main one), we need to figure out a solution before this can
        # be enabled again.
        return {}

    if not snapshot_id:
        return {}

    snapshot = snapshot_get_all_branches(storage, snapshot_id)
    if not snapshot:
        return {}

    branches = snapshot.branches.values()

    tip_revision_ids = []
    tip_release_ids = []

    for branch in branches:
        if branch.target_type == SnapshotTargetType.REVISION:
            tip_revision_ids.append(branch.target)
        elif branch.target_type == SnapshotTargetType.RELEASE:
            tip_release_ids.append(branch.target)

    revision_datetimes = [
        revision.date.to_datetime()
        for revision in storage.revision_get(tip_revision_ids)
        if revision and revision.date
    ]

    release_datetimes = [
        release.date.to_datetime()
        for release in storage.release_get(tip_release_ids)
        if release and release.date
    ]

    ret = {}
    if revision_datetimes:
        ret["last_revision_date"] = max(revision_datetimes).isoformat()
    if release_datetimes:
        ret["last_release_date"] = max(release_datetimes).isoformat()

    return ret


def process_journal_objects(messages, *, search, storage=None):
    """Worker function for `JournalClient.process(worker_fn)`, after
    currification of `scheduler` and `task_names`."""
    assert set(messages) <= EXPECTED_MESSAGE_TYPES, set(messages)

    if "origin" in messages:
        process_origins(messages["origin"], search)

    if "origin_visit_status" in messages:
        process_origin_visit_statuses(messages["origin_visit_status"], search, storage)

    if "origin_intrinsic_metadata" in messages:
        process_origin_intrinsic_metadata(messages["origin_intrinsic_metadata"], search)

    if "origin_extrinsic_metadata" in messages:
        process_origin_extrinsic_metadata(messages["origin_extrinsic_metadata"], search)


def process_origins(origins, search):
    logging.debug("processing origins %r", origins)

    search.origin_update(origins)


def process_origin_visit_statuses(visit_statuses, search, storage):
    logging.debug("processing origin visit statuses %r", visit_statuses)

    def hexify(b: Optional[bytes]) -> Optional[str]:
        if b is None:
            return None
        return b.hex()

    processed_visit_statuses = []
    for visit_status in visit_statuses:
        visit_types = []
        visit_type = visit_status.get("type")
        if visit_type is None:
            visit = storage.origin_visit_get_by(
                origin=visit_status["origin"], visit=visit_status["visit"]
            )
            if visit is not None:
                visit_types.append(visit.type)
        else:
            visit_types.append(visit_type)

        processed_status = {
            "url": visit_status["origin"],
            "visit_types": visit_types,
        }
        if visit_status["status"] == "full":
            processed_status.update(
                {
                    "has_visits": True,
                    "nb_visits": visit_status["visit"],
                    "snapshot_id": hexify(visit_status.get("snapshot")),
                    "last_visit_date": visit_status["date"].isoformat(),
                    "last_eventful_visit_date": visit_status["date"].isoformat(),
                    **fetch_last_revision_release_date(
                        visit_status.get("snapshot"), storage
                    ),
                }
            )
        processed_visit_statuses.append(processed_status)

    if processed_visit_statuses:
        search.origin_update(processed_visit_statuses)


def process_origin_intrinsic_metadata(origin_metadata, search):
    logging.debug("processing origin intrinsic_metadata %r", origin_metadata)

    origin_metadata = [
        {
            "url": item["id"],
            "jsonld": item["metadata"],
        }
        for item in origin_metadata
    ]

    search.origin_update(origin_metadata)


def process_origin_extrinsic_metadata(origin_metadata, search):
    logging.debug("processing origin extrinsic_metadata %r", origin_metadata)

    origin_metadata = [
        {
            "url": item["id"],
            "jsonld": item["metadata"],
        }
        for item in origin_metadata
    ]

    search.origin_update(origin_metadata)



#!/usr/bin/env python3
# Copyright (C) 2015-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import glob
import os
import shutil
import subprocess

from setuptools import Command, setup
from setuptools.command.build_py import build_py
from setuptools.command.develop import develop
from setuptools.command.sdist import sdist


def needs_regen(dest, sources) -> bool:
    """Returns whether any of the 'sources' files was modified after 'dst'."""
    if not os.path.exists(dest):
        return True

    dest_mtime = os.stat(dest).st_mtime

    for source in sources:
        if os.stat(source).st_mtime > dest_mtime:
            return True

    return False


yarn = os.environ.get("YARN", "yarnpkg" if shutil.which("yarnpkg") else "yarn")


class TSCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass


class TSInstallCommand(TSCommand):
    description = "Installs node_modules related to query language"

    def run(self):
        subprocess.run([yarn, "install"], check=True)


class TSBuildSoCommand(TSCommand):
    description = "Builds swh_ql.so"

    def initialize_options(self):
        self.build_lib = None
        super().initialize_options()

    def finalize_options(self):
        self.set_undefined_options("build", ("build_lib", "build_lib"))
        super().finalize_options()

    def run(self):
        ql_dir = os.path.join(self.build_lib, "swh/search/query_language")
        copy_ql_tree(ql_dir)
        if needs_regen(
            os.path.join(ql_dir, "src/parser.c"),
            glob.glob("swh/search/query_language/**/*"),
        ):
            print("parser.c missing from build dir.")
            self.run_command("ts_install")
            generate_parser(ql_dir)

        static_dir = os.path.join(self.build_lib, "swh/search/static")
        os.makedirs(static_dir, exist_ok=True)

        # This import cannot be toplevel, as setuptools installs it after the script
        # starts running
        from tree_sitter import Language

        Language.build_library(os.path.join(static_dir, "swh_ql.so"), [ql_dir])
        print("swh_ql.so file generated")


class TSBuildCommand(TSCommand):
    description = "Builds swh_ql.so"

    def run(self):
        self.run_command("ts_build_so")


class custom_build(build_py):
    def run(self):
        super().run()
        if not self.dry_run:
            if getattr(self, "editable_mode", False):
                # setuptools >= 64
                self.run_command("ts_install")
                generate_parser("swh/search/query_language")
            else:
                self.run_command("ts_build")

    def get_output_mapping(self):
        output_mapping = super().get_output_mapping()
        for fname in (
            "swh/search/query_language/src/grammar.json",
            "swh/search/query_language/src/tree_sitter/parser.h",
            "swh/search/query_language/src/parser.c",
            "swh/search/query_language/src/node-types.json",
        ):
            output_mapping[f"{self.build_lib}/{fname}"] = fname
        return output_mapping

    def get_source_files(self):
        source_files = super().get_source_files()
        source_files.extend(
            [
                "swh/search/query_language/grammar.js",
                "swh/search/query_language/token.js",
            ]
        )
        return source_files


class custom_sdist(sdist):
    def make_release_tree(self, base_dir, files):
        super().make_release_tree(base_dir, files)

        dist_ql_path = os.path.join(base_dir, "swh/search/query_language")

        if not self.dry_run:
            self.run_command("ts_install")

            copy_ql_tree(dist_ql_path)
            generate_parser(dist_ql_path)


class custom_develop(develop):
    def run(self):
        super().run()
        if not self.dry_run:
            self.run_command("ts_install")
            generate_parser("swh/search/query_language")


def copy_ql_tree(dest_path):
    # FIXME: setuptools should copy this itself...
    print("Copying parser files")
    if os.path.exists(dest_path):
        shutil.rmtree(dest_path)
    shutil.copytree("swh/search/query_language", dest_path)


def generate_parser(dest_path):
    print("Getting path")
    path = subprocess.check_output([yarn, "bin"]).decode().strip()
    env = {**os.environ, "PATH": os.pathsep.join([path, os.environ["PATH"]])}
    print("Generating")
    subprocess.run(["tree-sitter", "generate", "--no-bindings"], cwd=dest_path, env=env)


setup(
    cmdclass={
        "build_py": custom_build,
        "sdist": custom_sdist,
        "develop": custom_develop,
        "ts_install": TSInstallCommand,
        "ts_build_so": TSBuildSoCommand,
        "ts_build": TSBuildCommand,
    },
)



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import codecs
from datetime import datetime
from typing import Optional

import iso8601


def get_expansion(field, sep=None):
    METADATA_FIELDS = {
        "licenses": ["jsonld", "http://schema.org/license", "@id"],
        "programming_languages": [
            "jsonld",
            "http://schema.org/programmingLanguage",
            "@value",
        ],
        "keywords": [
            "jsonld",
            "http://schema.org/keywords",
            "@value",
        ],
        "descriptions": [
            "jsonld",
            "http://schema.org/description",
            "@value",
        ],
        "date_created": [
            "jsonld",
            "http://schema.org/dateCreated",
            "@value",
        ],
        "date_modified": [
            "jsonld",
            "http://schema.org/dateModified",
            "@value",
        ],
        "date_published": [
            "jsonld",
            "http://schema.org/datePublished",
            "@value",
        ],
    }

    if sep:
        return sep.join(METADATA_FIELDS[field])

    return METADATA_FIELDS[field]


def parse_and_format_date(date_str: str) -> Optional[str]:
    """
    Parses a string date in the format %Y-%m-%d or ISO8601 and returns
    a new string date in the format YYYY-mm-dd if the parsing succeeded
    otherwise None.
    """
    try:
        return datetime.strptime(date_str, "%Y-%m-%d").strftime("%Y-%m-%d")
    except Exception:
        try:
            return iso8601.parse_date(date_str).strftime("%Y-%m-%d")
        except Exception:
            return None


def escape(obj):
    r"""Makes the object directly injectable into the
    query language by converting the escapable parts of
    the object into escape sequences.

    For strings, appends \ before special characters like ', ", and \

    For arrays, applies the same transformation on each element, joins the
    elements and returns a string-like representation of the list.

    >>> print(escape("foo ' bar"))
    "foo \' bar"

    >>> print(escape([r"foo ' bar", r"bar \\\' baz", r'foo " baz']))
    ["foo \' bar", "bar \\\\\\\' baz", "foo \" baz"]

    """
    if type(obj) is list:
        items = [escape(item) for item in obj]
        return "[" + ", ".join(items) + "]"
    elif type(obj) is str:
        return (
            '"'
            + obj.translate(
                {
                    ord("'"): r"\'",
                    ord('"'): r"\"",
                    ord("\\"): r"\\",
                }
            )
            + '"'
        )
    else:
        raise Exception(f"Unexpected item type {type(obj)}")


def unescape(string):
    r"""Processes the escaped special characters

    >>> unescape(r'''foo " bar''') == r'''foo " bar'''
    True
    >>> unescape(r'''foo \" bar''') == r'''foo " bar'''
    True
    >>> unescape(r'''foo \\" bar''') == r'''foo \" bar'''
    True
    >>> unescape(r'''foo \\\" bar''') == r'''foo \" bar'''
    True
    >>> unescape(r'''foo \\\\" bar''') == r'''foo \\" bar'''
    True
    >>> unescape(r'''café \" foo''') == r'''café " foo'''
    True
    """
    return codecs.escape_decode(string.encode())[0].decode()
